#!/usr/bin/env python
# Copyright (c) 2018, Clemence Frioux <clemence.frioux@inria.fr>
#
# This file is part of miscoto.
#
# miscoto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# miscoto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miscoto.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-

import argparse
import sys
import os
import time
import logging
from miscoto import query, sbml, commons, utils
from os import listdir
from os.path import isfile, join
from pyasp.asp import *
from xml.etree.ElementTree import ParseError

logger = logging.getLogger(__name__)

###############################################################################
#
message = """
Compute the scope and target produciblity of a host (optional) and the added-value
of a microbiome regarding scope and target producibility. The microbiome result part
gives the targets and compounds that are producible providing cooperation occurs
within the community of host + all symbionts and that were not producible with
the host alone.
Computation from SBML models or an instance pre-created with miscoto_instance.py
"""

requires = """
requires PyASP package: "pip install PyASP"
"""

pusage="""
**1** from SBML2 files \n
python miscoto_scopes.py -m host.sbml -b symbiont_directory -s seeds.sbml -t targets.sbml
\n
**2** from SBML2 files \n
python miscoto_scopes.py -b symbiont_directory -s seeds.sbml -t targets.sbml
\n
**3** from a pre-computed instance with possibly (additional) seeds or targets \n
python miscoto_scopes.py -a instance.lp [-s seeds.sbml] [-t targets.sbml]
"""
#
###############################################################################


def cmd_scopes():
    """run directly miscoto_scopes from the shell
    """
    parser = argparse.ArgumentParser(description=message, usage=pusage, epilog=requires)
    parser.add_argument("-a", "--asp",
                        help="instance if already created with miscoto_instance", required=False)
    parser.add_argument("-m", "--modelhost",
                        help="host model in SBML format, ignored if -a instance is provided",
                        required=False)
    parser.add_argument("-b", "--bactsymbionts",
                        help="directory of symbionts models, all in sbml format, ignored if -a instance is provided",
                        required=False)
    parser.add_argument("-s", "--seeds",
                        help="seeds in SBML format",
                        required=False)
    parser.add_argument("-t", "--targets",
                        help="targets in SBML format",
                        required=False)

    args = parser.parse_args()
    lp_instance_file_arg = args.asp
    targets_sbml = args.targets
    seeds_sbml = args.seeds
    bacterium_met =  args.bactsymbionts
    draft_sbml = args.modelhost
    run_scopes(lp_instance_file_arg, targets_sbml, seeds_sbml, bacterium_met, draft_sbml)

def run_scopes(lp_instance_file=None, targets_file=None, seeds_file=None, bacteria_dir=None, host_file=None):
    """[summary]
        lp_instance_file ([str], optional): Defaults to None. [ASP facts instance of the problem]
        targets_file ([str], optional): Defaults to None. [targets file]
        seeds_file ([str], optional): Defaults to None. [seeds file]
        bacteria_dir ([str], optional): Defaults to None. [directory of bacterial metabolic networks]
        host_file ([str], optional): Defaults to None. [host metabolic network]
    
    Returns:
        [dic]: [all information related to scope computation]
    """

    start_time = time.time()
    results = {}
    # case 1: instance is provided, just read targets and seeds if given
    input_instance = False
    if lp_instance_file:
        if not os.path.isfile(lp_instance_file) :
            logger.info('Instance file not found')
            sys.exit(1)

        input_instance = True
        delete_lp_instance = False
        logger.info(
            "Instance provided, only seeds and targets will be added if given")

        if targets_file:
            logger.info('Reading targets from ' + targets_file)
            try:
                targetsfacts = sbml.readSBMLspecies(targets_file, 'target')
            except FileNotFoundError:
                logger.critical('Targets file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: "+targets_file)
                sys.exit(1)
        else:
            targetsfacts = TermSet()

        if seeds_file:
            logger.info('Reading targets from ' + seeds_file)
            try:
                seedsfacts = sbml.readSBMLspecies(seeds_file, 'seed')
            except FileNotFoundError:
                logger.critical('Seeds file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: " + seeds_file)
                sys.exit(1)
        else:
            seedsfacts = TermSet()

            with open(lp_instance_file, "a") as f:
                for elem in targetsfacts:
                    f.write(str(elem) + '.\n')
                for elem in seedsfacts:
                    f.write(str(elem) + '.\n')

    # case 2: read inputs from SBML files
    elif bacteria_dir and seeds_file:
        if not os.path.isdir(bacteria_dir):
            logger.info("Symbiont directory not found")
            sys.exit(1)

        delete_lp_instance = True

        if host_file:
            logger.info('Reading host network from ' + host_file)
            try:
                draftnet = sbml.readSBMLnetwork_symbionts(host_file, 'host_metab_mod')
            except FileNotFoundError:
                logger.critical('Host file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: " + host_file)
                sys.exit(1)
            draftnet.add(Term('draft', ["\"" + 'host_metab_mod' + "\""]))
        else:
            logger.warning('No host provided.')
            draftnet = TermSet()

        logger.info('Reading seeds from ' + seeds_file)
        try:
            seeds = sbml.readSBMLspecies(seeds_file, 'seed')
        except FileNotFoundError:
            logger.critical('Seeds file not found')
            sys.exit(1)
        except ParseError:
            logger.critical("Invalid syntax in SBML file: "+seeds_file)
            sys.exit(1)
        lp_instance = TermSet(draftnet.union(seeds))

        if targets_file:
            logger.info('Reading targets from ' + targets_file)
            try:
                targets = sbml.readSBMLspecies(targets_file, 'target')
            except FileNotFoundError:
                logger.critical('Targets file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: " + targets_file)
                sys.exit(1)
            lp_instance = TermSet(lp_instance.union(targets))
        else:
            logger.info("No targets provided.")

        if not os.path.isdir(bacteria_dir):
            logger.critical("Symbiont directory not found")
            sys.exit(1)

        lp_instance_file = utils.to_file(lp_instance)

        logger.info('Reading bacterial networks from ' + bacteria_dir + '...')
        bactfacts = TermSet()
        onlyfiles = [f for f in listdir(bacteria_dir) if isfile(join(bacteria_dir, f))]
        for bacteria_file in onlyfiles:
            name = os.path.splitext(bacteria_file)[0]
            try:
                one_bact_model = sbml.readSBMLnetwork_symbionts(bacteria_dir+'/'+bacteria_file, name)
                one_bact_model.add(Term('bacteria', ["\"" + name + "\""]))
                utils.to_file(one_bact_model, lp_instance_file)
                logger.info('Done for ' + name)
            except:
                logger.info('Could not read file ' + name + ', will ignore it')

    else:
        logger.critical("ERROR missing input")
        logger.info(pusage)
        quit()

    logger.info("Computing scopes...")
    try:
        model = query.get_scopes(lp_instance_file, commons.ASP_SRC_SCOPES)
    except OSError:
        logger.critical(
            "Error. Possible reason: solvers are not properly installed. Please install them again by running 'pip uninstall pyasp' and 'pip install pyasp --no-cache-dir'"
        )
        sys.exit(1)
    host_scope = []
    host_prodtargets = []
    host_unprodtargets = []
    com_scope = []
    comhost_scope = []
    com_prodtargets = []
    com_unprodtargets = []
    for a in model:
        if a.pred() == 'dscope':
            host_scope.append(a.arg(0))
        elif a.pred() == 'dproducible':
            host_prodtargets.append(a.arg(0))
        elif a.pred() == 'dunproducible':
            host_unprodtargets.append(a.arg(0))
        elif a.pred() == 'newscope_microbiome':
            com_scope.append(a.arg(0))
        elif a.pred() == 'newscope_with_host':
            comhost_scope.append(a.arg(0))
        elif a.pred() == 'newlyproducible':
            com_prodtargets.append(a.arg(0))
        elif a.pred() == 'aunproducible':
            com_unprodtargets.append(a.arg(0))

    if host_file or input_instance:
        logger.info('*** HOST model producibility check ***')

        logger.info('Host producible targets => ' + str(len(host_prodtargets)))
        logger.info("\n".join(host_prodtargets))
        results['host_prodtargets'] = host_prodtargets

        logger.info('Host unproducible targets => ' +
                    str(len(host_unprodtargets)))
        logger.info("\n".join(host_unprodtargets))
        results['host_unprodtargets'] = host_unprodtargets

        logger.info('Host scope => ' + str(len(host_scope)))
        logger.info("\n".join(host_scope))
        results['host_scope'] = host_scope

    logger.info('*** MICROBIOME added-value ***')
    logger.info('Microbiome only producible targets => ' +
                str(len(com_prodtargets)))
    logger.info("\n".join(com_prodtargets))
    results['com_prodtargets'] = com_prodtargets

    logger.info('Microbiome unproducible targets => ' +
                str(len(com_unprodtargets)))
    logger.info("\n".join(com_unprodtargets))
    results['com_unprodtargets'] = com_unprodtargets

    if host_file or input_instance:
        logger.info(
            'Microbiome only (host + symbionts) scope (host metabolites only producible with the microbiome) => '
            + str(len(com_scope)))
        logger.info("\n".join(comhost_scope))
        results['comhost_scope'] = comhost_scope
    if input_instance or not host_file:
        logger.info(
            'Microbiome only (symbionts) scope (metabolites only producible with the microbiome) => '
            + str(len(com_scope)))
        logger.info("\n".join(com_scope))
        results['com_scope'] = com_scope

    if delete_lp_instance == True:
        os.unlink(lp_instance_file)

    logger.info("--- %s seconds ---" % (time.time() - start_time))
    utils.clean_up()
    return results

if __name__ == '__main__':
    cmd_scopes()


#!/usr/bin/python
# -*- coding: utf-8 -*-
import re
from pyasp.asp import *
from pyasp.term import *
import xml.etree.ElementTree as etree
from xml.etree.ElementTree import XML, fromstring, tostring

def get_model(sbml):
    """Get the model of a SBML

    Args:
        sbml (str): SBML file

    Returns:
        xml.etree.ElementTree.Element: SBML model
    """
    model_element = None
    for e in sbml:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "model":
            model_element = e
            break
    return model_element

def get_listOfCompartments(model):
    """Get list of compartments in a SBML model
    
    Args:
        model (xml.etree.ElementTree.Element): SBML model
    
    Returns:
        xml.etree.ElementTree.Element: list of compartments
    """
    listOfCompartments = None
    for e in model:
        if e.tag[0] == "{":
          uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "listOfCompartments":
          listOfCompartments = e
          break
    return listOfCompartments

def get_listOfSpecies(model):
    """Get list of Species of a SBML model
    
    Args:
        model (xml.etree.ElementTree.Element): SBML model
    
    Returns:
        xml.etree.ElementTree.Element: list of species
    """
    listOfSpecies = None
    for e in model:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "listOfSpecies":
            listOfSpecies = e
            break
    return listOfSpecies

def get_listOfReactions(model):
    """Get list of reactions of a SBML model
    
    Args:
        model (xml.etree.ElementTree.Element): SBML model
    
    Returns:
        xml.etree.ElementTree.Element: list of reactions
    """
    listOfReactions = None
    for e in model:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "listOfReactions":
            listOfReactions = e
            break
    return listOfReactions

def get_listOfReactants(reaction):
    """Get list of reactants of a reaction
    
    Args:
        reaction (xml.etree.ElementTree.Element): reaction
    
    Returns:
        xml.etree.ElementTree.Element: list of reactants
    """
    listOfReactants = None
    for e in reaction:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "listOfReactants":
            listOfReactants = e
            break
    return listOfReactants

def get_listOfProducts(reaction):
    """Get list of porducts of a reaction
    
    Args:
        reaction (xml.etree.ElementTree.Element): reaction
    
    Returns:
        xml.etree.ElementTree.Element: list of products
    """
    listOfProducts = None
    for e in reaction:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "listOfProducts":
            listOfProducts = e
            break
    return listOfProducts

def readSBMLnetwork_symbionts(filename, name) :
    """Read a SBML metabolic network
    
    Args:
        filename (str): SBML file
        name (str): suffix to identify the type of network
    
    Returns:
        TermSet: metabolic model
    """
    lpfacts = TermSet()

    tree = etree.parse(filename)
    sbml = tree.getroot()
    model = get_model(sbml)

    listOfReactions = get_listOfReactions(model)
    for e in listOfReactions:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "reaction":
            reactionId = e.attrib.get("id")
            lpfacts.add(Term('reaction', ["\""+reactionId+"\"", "\""+name+"\""])) 
            if(e.attrib.get("reversible")=="true"):
                lpfacts.add(Term('reversible', ["\""+reactionId+"\"", "\""+name+"\""]))

            listOfReactants = get_listOfReactants(e)
            # if listOfReactants == None :
            #     print("\n Warning:" + reactionId + "listOfReactants=None")
            if listOfReactants != None:
                for r in listOfReactants:
                    lpfacts.add(Term('reactant', ["\""+r.attrib.get("species").replace('"','')+"\"", "\""+reactionId+"\"", "\""+name+"\""])) #,"\""+name+"\""

            listOfProducts = get_listOfProducts(e)
            # if listOfProducts == None:
            #     print("\n Warning:" + reactionId + "listOfProducts=None")
            if listOfProducts != None:
                for p in listOfProducts:
                    lpfacts.add(Term('product', ["\""+p.attrib.get("species").replace('"','')+"\"", "\""+reactionId+"\"", "\""+name+"\""])) #,"\""+name+"\""
    listofspecies = get_listOfSpecies(model)
    for e in listofspecies:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "species":
            speciesId = e.attrib.get("id").replace('"','')
            speciesNm = e.attrib.get("name").replace('"','')
            compartment = e.attrib.get("compartment")
            lpfacts.add(Term('species', ["\""+speciesId+"\"", "\""+speciesNm+"\"", "\""+compartment+"\"", "\""+name+"\""]))

    return lpfacts

def readSBMLnetwork_symbionts_noemptyrctprd(filename, name) :
    """Read a SBML metabolic network while ignoring the reactions that have no reactant or product
    
    Args:
        filename (str): SBML file
        name (str): suffix to identify the type of network
    
    Returns:
        TermSet: metabolic model
    """
    lpfacts = TermSet()

    tree = etree.parse(filename)
    sbml = tree.getroot()
    model = get_model(sbml)

    listOfReactions = get_listOfReactions(model)
    for e in listOfReactions:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "reaction":
            reactionId = e.attrib.get("id")

            listOfReactants = get_listOfReactants(e)
            listOfProducts = get_listOfProducts(e)
            if listOfReactants != None and listOfProducts != None:
                lpfacts.add(Term('reaction', ["\""+reactionId+"\"", "\""+name+"\""]))
                if e.attrib.get("reversible")=="true":
                    lpfacts.add(Term('reversible', ["\""+reactionId+"\"", "\""+name+"\""]))
                for r in listOfReactants:
                    lpfacts.add(Term('reactant', ["\""+r.attrib.get("species")+"\"", "\""+reactionId+"\"", "\""+name+"\""])) #,"\""+name+"\""
                for p in listOfProducts:
                    lpfacts.add(Term('product', ["\""+p.attrib.get("species")+"\"", "\""+reactionId+"\"", "\""+name+"\""])) #,"\""+name+"\""
            else:
                print("reaction " + reactionId + " was ignored due to absent reactants or products")
    listofspecies = get_listOfSpecies(model)
    for e in listofspecies:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "species":
            speciesId = e.attrib.get("id")
            speciesNm = e.attrib.get("name")
            compartment = e.attrib.get("compartment")
            lpfacts.add(Term('species', ["\""+speciesId+"\"", "\""+speciesNm+"\"", "\""+compartment+"\"", "\""+name+"\""]))

    #print(lpfacts)
    return lpfacts

def readSBMLnetwork_em(filename, externalcomp="e") :
    """Read a SBML metabolic network while considering external compartment
    
    Args:
        filename (str): SBML file
        externalcomp (str): Default to 'e'. external compartment
    
    Returns:
        TermSet, str: metabolic model and its name
    """
    lpfacts = TermSet()

    tree = etree.parse(filename)
    sbml = tree.getroot()
    model = get_model(sbml)

    name = model.attrib.get("id")
    if name == None:
        print('Error: target and network smbl files for one organism must have identical model IDs. ' + filename + ' network file misses a model ID')
        quit()

    listOfReactions = get_listOfReactions(model)
    for e in listOfReactions:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "reaction":
            reactionId = e.attrib.get("id")
            lpfacts.add(Term('dreaction', ["\""+reactionId+"\"", "\""+name+"\""])) #
            if(e.attrib.get("reversible")=="true"):  lpfacts.add(Term('reversible', ["\""+reactionId+"\"", "\""+name+"\""]))

            listOfReactants = get_listOfReactants(e)
            # if listOfReactants== None : print("\n Warning:" + reactionId, " listOfReactants=None")
            #else:
            if listOfReactants != None:
                for r in listOfReactants:
                    lpfacts.add(Term('reactant', ["\""+r.attrib.get("species")+"\"", "\""+reactionId+"\"","\""+name+"\""])) #

            listOfProducts = get_listOfProducts(e)
            # if listOfProducts== None : print("\n Warning:" + reactionId, " listOfProducts=None")
            #else:
            if listOfProducts != None:
                for p in listOfProducts:
                    lpfacts.add(Term('product', ["\""+p.attrib.get("species")+"\"", "\""+reactionId+"\"","\""+name+"\""])) #
    lpfacts.add(Term('organism', ["\"" + name + "\""]))
    lpfacts.add(Term('external', ["\"" + externalcomp + "\""]))

    listofspecies = get_listOfSpecies(model)
    for e in listofspecies:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "species":
            speciesId = e.attrib.get("id")
            speciesNm = e.attrib.get("name")
            compartment = e.attrib.get("compartment")
            lpfacts.add(Term('species', ["\""+speciesId+"\"", "\""+speciesNm+"\"", "\""+compartment+"\"", "\""+name+"\""]))

    listofcompartments = get_listOfCompartments(model)
    for e in listofcompartments:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "compartment":
            compartmentId = e.attrib.get("id")
            lpfacts.add(Term('compartment', ["\""+compartmentId+"\"", "\""+name+"\""]))
    #print(lpfacts)
    return lpfacts, name

def readSBMLnetwork_em_noemptyrctprd(filename, externalcomp="e") :
    """Read a SBML metabolic network while ignoring the reactions that have no reactant or product
    
    Args:
        filename (str): SBML file
        externalcomp (str): Default to 'e'. external compartment
    
    Returns:
        TermSet, str: metabolic model and its name
    """
    lpfacts = TermSet()

    tree = etree.parse(filename)
    sbml = tree.getroot()
    model = get_model(sbml)

    name = model.attrib.get("id")
    if name == None:
        print('Error: target and network smbl files for one organism must have identical model IDs. ' + filename + ' network file misses a model ID')
        quit()

    listOfReactions = get_listOfReactions(model)
    for e in listOfReactions:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "reaction":
            reactionId = e.attrib.get("id")

            listOfReactants = get_listOfReactants(e)
            listOfProducts = get_listOfProducts(e)
            # if listOfReactants == None :
            #     print("\n Warning:" + reactionId + "listOfReactants=None")
            if listOfReactants != None and listOfProducts != None:
                lpfacts.add(Term('dreaction', ["\""+reactionId+"\"", "\""+name+"\""])) #
                if(e.attrib.get("reversible")=="true"):  lpfacts.add(Term('reversible', ["\""+reactionId+"\"", "\""+name+"\""]))
                if e.attrib.get("reversible")=="true":
                    lpfacts.add(Term('reversible', ["\""+reactionId+"\"", "\""+name+"\""]))
                for r in listOfReactants:
                    lpfacts.add(Term('reactant', ["\""+r.attrib.get("species")+"\"", "\""+reactionId+"\"", "\""+name+"\""])) #,"\""+name+"\""
                for p in listOfProducts:
                    lpfacts.add(Term('product', ["\""+p.attrib.get("species")+"\"", "\""+reactionId+"\"", "\""+name+"\""])) #,"\""+name+"\""
            else:
                print("reaction " + reactionId + " was ignored due to absent reactants or products ( org=" + name +' )')

    lpfacts.add(Term('organism', ["\"" + name + "\""]))
    lpfacts.add(Term('external', ["\"" + externalcomp + "\""]))

    listofspecies = get_listOfSpecies(model)
    for e in listofspecies:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "species":
            speciesId = e.attrib.get("id")
            speciesNm = e.attrib.get("name")
            compartment = e.attrib.get("compartment")
            lpfacts.add(Term('species', ["\""+speciesId+"\"", "\""+speciesNm+"\"", "\""+compartment+"\"", "\""+name+"\""]))

    listofcompartments = get_listOfCompartments(model)
    for e in listofcompartments:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "compartment":
            compartmentId = e.attrib.get("id")
            lpfacts.add(Term('compartment', ["\""+compartmentId+"\"", "\""+name+"\""]))
    return lpfacts, name

def readSBMLtargets(filename):
    """Get SBML targets
    
    Args:
        filename (str): SBML file
    
    Returns:
        TermSet, str: targets and name of the model
    """
    lpfacts = TermSet()

    tree = etree.parse(filename)
    sbml = tree.getroot()
    model = get_model(sbml)

    name = model.attrib.get("id")
    if name == None:
        print('Error: target and network smbl files for one organism must have identical model IDs. ' + filename + ' target file misses a model ID')
        quit()

    listOfSpecies = get_listOfSpecies(model)
    for e in listOfSpecies:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else: tag = e.tag
        if tag == "species":
            lpfacts.add(Term('target', ["\""+e.attrib.get("id")+"\"", "\""+name+"\""]))
    return lpfacts, name

def readSBMLspecies(filename, speciestype) :
    """Get species of a SBML (seeds, targets)
    
    Args:
        filename (str): SBML file
        speciestype (str): seed or target
    
    Returns:
        TermSet: seeds or targets
    """
    lpfacts = TermSet()

    tree = etree.parse(filename)
    sbml = tree.getroot()
    model = get_model(sbml)

    listOfSpecies = get_listOfSpecies(model)
    for e in listOfSpecies:
        if e.tag[0] == "{":
            uri, tag = e.tag[1:].split("}")
        else:
            tag = e.tag
        if tag == "species":
            lpfacts.add(Term(speciestype, ["\""+e.attrib.get("id").replace('"','')+"\""]))
    return lpfacts



from miscoto.miscoto_instance import cmd_instance
from miscoto.miscoto_mincom import cmd_mincom
from miscoto.miscoto_scopes import cmd_scopes
import logging

logging.basicConfig(format='%(message)s', level=logging.DEBUG)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def main_instance(args=None):
    cmd_instance()

def main_mincom(args=None):
    cmd_mincom()

def main_scopes(args=None):
    cmd_scopes()



import os
import tempfile
from pyasp.asp import *

def get_scopes(instance_f, encoding):
    """Get metabolic scope of a microbiota
    
    Args:
        instance_f (str): ASP instance file
        encoding (str): ASP model encoding
    
    Returns:
        TermSet: ASP model
    """
    prg = [encoding, instance_f]
    options = ''
    solver = Gringo4Clasp(clasp_options=options)
    models = solver.run(prg,collapseTerms=True,collapseAtoms=False)
    return models[0]


def get_grounded_communities(instance, encoding):
    """Ground the model, from a TermSet
    
    Args:
        instance (TermSet): model
        encoding (str): ASP model encoding
    
    Returns:
        bytes: grounded model
    """
    instance_f = instance.to_file()
    # print(os.path.abspath(instance_f))
    prg = [encoding, instance_f]
    grounder = Gringo4()
    grounding = grounder.run(prg)
    os.unlink(instance_f)
    return grounding

def get_grounded_communities_from_file(instance_f, encoding):
    """Ground the model, from a file
    
    Args:
        instance_f (str): model file
        encoding (str): ASP model encoding
    
    Returns:
        bytes: grounded model
    """
    prg = [encoding, instance_f]
    grounder = Gringo4()
    grounding = grounder.run(prg)
    # os.unlink(instance_f)
    return grounding

def get_communities_from_g(grounding):
    """Get optimal community, from grounding
    
    Args:
        grounding (bytes): grounded model
    
    Returns:
        TermSet: solution
    """
    options = '--configuration jumpy --opt-strategy=usc,5'
    solver = Clasp(clasp_options=options)
    models = solver.run(grounding,collapseTerms=True,collapseAtoms=False)
    return models[0]

def get_communities(lp_instance, encoding):
    """Get optimial community, from TermSet
    
    Args:
        lp_instance (TermSet): microbiota model
        encoding (str): ASP model encoding
    
    Returns:
        TermSet: solution
    """
    lp_f = lp_instance.to_file()
    prg = [encoding, lp_f]
    print(os.path.abspath(lp_f))
    options = '--configuration jumpy --opt-strategy=usc,5'
    solver = Gringo4Clasp(clasp_options=options)
    models = solver.run(prg,collapseTerms=True,collapseAtoms=False)
    os.unlink(lp_f)
    # print(models[0])
    return models[0]

def get_intersection_communities_from_g(grounding, optimum):
    """Get intersection of solutions, from grounding
    
    Args:
        grounding (bytes): grounded model
        optimum (str): optimal score
    
    Returns:
        TermSet: intersection
    """
    options = '--configuration jumpy --opt-strategy=usc,5 --enum-mode cautious --opt-mode=optN --opt-bound=' +str(optimum)
    solver = Clasp(clasp_options=options)
    intersec = solver.run(grounding,collapseTerms=True,collapseAtoms=False)
    return intersec[0]

def get_intersection_communities_from_g_noopti(grounding):
    """Get intersection of solutions, from grounding, without optimal score
    
    Args:
        grounding (bytes): grounded model
    
    Returns:
        TermSet: intersection
    """
    options = '--configuration jumpy --opt-strategy=usc,5 --enum-mode cautious --opt-mode=optN'
    solver = Clasp(clasp_options=options)
    intersec = solver.run(grounding,collapseTerms=True,collapseAtoms=False)
    return intersec[0]

def get_intersection_communities(lp_instance, optimum, encoding):
    """Get intersection of solutions, from TermSet
    
    Args:
        lp_instance (TermSet): microbiota model
        optimum (str): optimal score
        encoding (str): ASP model encoding
    
    Returns:
        TermSet: intersection
    """
    lp_f = lp_instance.to_file()
    prg = [encoding, lp_f]
    options = '--configuration jumpy --opt-strategy=usc,5 --enum-mode cautious --opt-mode=optN --opt-bound=' +str(optimum)
    solver = Gringo4Clasp(clasp_options=options)
    intersec = solver.run(prg,collapseTerms=True,collapseAtoms=False)
    os.unlink(lp_f)
    return intersec[0]

def get_all_communities_from_g(grounding, optimum, nmodels=0):
    """Get all optimal communities, from grounding
    
    Args:
        grounding (bytes): grounded model
        optimum (str): optimal score
        nmodels (int, optional): Defaults to 0. number of models to compute, 0 = all
    
    Returns:
        list: list of Termsets
    """
    options = str(nmodels)+' --configuration handy --opt-strategy=usc,5 --opt-mode=optN --opt-bound=' +str(optimum)
    solver = Clasp(clasp_options=options)
    models = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return models

def get_all_communities_from_g_noopti(grounding, nmodels=0):
    """Get all optimal communities, from grounding, without optimal score
    
    Args:
        grounding (bytes): grounded model
        nmodels (int, optional): Defaults to 0. number of models, 0 = all
    
    Returns:
        list: list of TermSets
    """
    options = str(nmodels)+' --configuration handy --opt-strategy=usc,5 --opt-mode=optN '
    solver = Clasp(clasp_options=options)
    models = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return models

def get_all_communities(lp_instance, optimum, encoding, nmodels=0):
    """Get all communities, from TermSet
    
    Args:
        lp_instance (TermSet): microbiota model
        optimum (str): optimal score
        encoding (str): ASP model encoding file
        nmodels (int, optional): Defaults to 0. number of models, 0 = all
    
    Returns:
        list: list of TermSets
    """
    lp_f = lp_instance.to_file()
    prg = [encoding, lp_f]
    options = str(nmodels)+' --configuration handy --opt-strategy=usc,5 --opt-mode=optN --opt-bound=' +str(optimum)
    solver = Gringo4Clasp(clasp_options=options)
    models = solver.run(prg, collapseTerms=True, collapseAtoms=False)
    os.unlink(lp_f)
    return models

def get_union_communities_from_g(grounding, optimum):
    """Get union of all community solutions
    
    Args:
        grounding (bytes): grounded model
        optimum (str): optimal score
    
    Returns:
        TermSet: union
    """
    options ='--configuration jumpy --opt-strategy=usc,5 --enum-mode=brave --opt-mode=optN --opt-bound='+str(optimum)
    solver = Clasp(clasp_options=options)
    union = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return union[0]

def get_union_communities_from_g_noopti(grounding):
    """Get union of all community solutions, from grounding, without optimal score
    
    Args:
        grounding (bytes): grounded instance
    
    Returns:
        TermSet: union
    """
    options ='--configuration jumpy --opt-strategy=usc,5 --enum-mode brave --opt-mode=optN'
    solver = Clasp(clasp_options=options)
    union = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return union[0]

def get_union_communities(lp_instance, optimum, encoding):
    """Get union of community solutions, from TermSet
    
    Args:
        lp_instance (TermSet): microbiota model
        optimum (str): optimal score
        encoding (str): ASP encoding model file
    
    Returns:
        TermSet: union
    """
    lp_f = lp_instance.to_file()
    prg = [encoding, lp_f]
    options ='--configuration jumpy --opt-strategy=usc,5 --enum-mode=brave --opt-mode=optN --opt-bound='+str(optimum)
    solver = Gringo4Clasp(clasp_options=options)
    union = solver.run(prg, collapseTerms=True, collapseAtoms=False)
    os.unlink(lp_f)
    return union[0]

def get_unproducible(draft, seeds, targets, encoding):
    """Get unproducible targets in a microbiota
    
    Args:
        draft (TermSet): metabolic model
        seeds (TermSet): seeds
        targets (TermSet): targets
        encoding (str): ASP model encoding
    
    Returns:
        TermSet: unproducible targets
    """
    draft_f = draft.to_file()
    seed_f =  seeds.to_file()
    target_f = targets.to_file()
    prg = [encoding, draft_f, seed_f, target_f]
    solver = Gringo4Clasp()
    models = solver.run(prg,collapseTerms=True,collapseAtoms=False)
    os.unlink(draft_f)
    os.unlink(seed_f)
    os.unlink(target_f)
    return models[0]

def get_transported(instance, encoding):
    """Get transported metabolites
    
    Args:
        instance (TermSet): microbiota model
        encoding (str): ASP model encoding
    
    Returns:
        TermSet: transported metabolites
    """
    instance_f = instance.to_file()
    prg = [encoding, instance_f]
    solver = Gringo4Clasp()
    models = solver.run(prg,collapseTerms=True,collapseAtoms=False)
    # print(os.path.abspath(instance_f))
    os.unlink(instance_f)
    return models[0]


def get_grounded_instance_exchanged_metabolites(instance, encoding, exchanged_in_escope=False):
    """Get grounding under compartmentalized framework
    
    Args:
        instance (TermSet): microbiota model
        encoding (str): ASP model encoding
        exchanged_in_escope (bool, default=False): additional option for ASP
    
    Returns:
        bytes: grounded model
    """
    instance_f = instance.to_file()
    if exchanged_in_escope:
        options = "--const exchanged_in_escope=1"
    else:
        options = "--const exchanged_in_escope=0"
    print(os.path.abspath(instance_f))
    prg = [encoding, instance_f]
    grounder = Gringo4(gringo_options=options)
    grounding = grounder.run(prg)
    os.unlink(instance_f)
    return grounding

def get_grounded_instance_exchanged_metabolites_from_file(instance_f, encoding, exchanged_in_escope=False):
    """Get grounding from file
    
    Args:
        instance_f (str): microbiota model file
        encoding (str): ASP model encoding
        exchanged_in_escope (bool, default=False): additional ASP option
    
    Returns:
        bytes: grounded model
    """
    if exchanged_in_escope:
        options = "--const exchanged_in_escope=1"
    else:
        options = "--const exchanged_in_escope=0"
    print(os.path.abspath(instance_f))
    prg = [encoding, instance_f]
    grounder = Gringo4(gringo_options=options)
    grounding = grounder.run(prg)
    return grounding

def get_exchanged_metabolites_onesol(grounding):
    """Select community in compartmentalized framework
    
    Args:
        grounding (bytes): grounded model
    
    Returns:
        TermSet: solution
    """
    options = "--configuration=jumpy --opt-strategy=usc,5"
    solver = Clasp(clasp_options=options)
    models = solver.run(grounding,collapseTerms=True,collapseAtoms=False)
    return models

def get_exchanged_metabolites_intersection(grounding, optimum):
    """Get intersection of communities solutions
    
    Args:
        grounding (bytes): grounded model
        optimum (str): optimal score
    
    Returns:
        TermSet: intersection
    """
    options='--configuration jumpy --opt-strategy=usc,5 --enum-mode cautious --opt-mode=optN --opt-bound='+str(optimum)
    solver = Clasp(clasp_options=options)
    intersec = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return intersec[0]

def get_exchanged_metabolites_allsol(grounding, optimum, nmodels=0):
    """Get all communities
    
    Args:
        grounding (bytes): grounded model
        optimum (str): optimal score
        nmodels (int, optional): Defaults to 0. number of models, 0 = all
    
    Returns:
        Set: set of TermSets
    """
    options = str(nmodels)+' --configuration jumpy --opt-strategy=usc,5 --opt-mode=optN --opt-bound='+str(optimum)
    solver = Clasp(clasp_options=options)
    models = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return models

def get_exchanged_metabolites_union(grounding, optimum):
    """Get union of community solutions
    
    Args:
        grounding (bytes): grounded model
        optimum (str): optimal score
    
    Returns:
        TermSet: union
    """
    options ='--configuration jumpy --opt-strategy=usc,5 --enum-mode brave --opt-mode=optN --opt-bound='+str(optimum)
    solver = Clasp(clasp_options=options)
    union = solver.run(grounding, collapseTerms=True, collapseAtoms=False)
    return union[0]



#!/usr/bin/env python
# Copyright (c) 2018, Clemence Frioux <clemence.frioux@inria.fr>
#
# This file is part of miscoto.
#
# miscoto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# miscoto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miscoto.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name             = 'Miscoto',
    version          = '1.1.5',
    url              = 'https://github.com/cfrioux/miscoto',
    license          = 'GPLv3+',
    description      = 'Microbiome Screening and COmmunity selection using TOpology',
    long_description = 'MiSCoTo is a Python3 tool to explore microbiomes and select minimal communities within them. It uses Answer Set Programming (ASP) to optimize community selection. \
Inputs: metabolic models, seeds (growth medium) and metabolic targets. \
Computations can be performed with a set of symbionts or a set of symbionts and a host. In the latter case, targets will be produced by the host, whereas in the former they will be produced by any member of the microbiome. \
More information on usage and troubleshooting on Github: https://github.com/cfrioux/miscoto',
    author           = 'Clemence Frioux',
    author_email     = 'clemence.frioux@gmail.com',
    packages         = ['miscoto'],
    package_dir      = {'miscoto' : 'miscoto'},
    package_data     = {'miscoto' : ['encodings/*.lp']},
    #scripts          = ['miscoto/miscoto_instance.py','miscoto/miscoto_mincom.py','miscoto/miscoto_scopes.py'],
    entry_points     = {'console_scripts': ['miscoto_instance = miscoto.__main__:main_instance', 'miscoto_mincom = miscoto.__main__:main_mincom',
                                            'miscoto_scopes = miscoto.__main__:main_scopes']},
    install_requires = ['pyasp == 1.4.3']
)



#!/usr/bin/env python
# Copyright (c) 2018, Clemence Frioux <clemence.frioux@inria.fr>
#
# This file is part of miscoto.
#
# miscoto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# miscoto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miscoto.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-

import argparse
import sys
import os
import time
import logging
from miscoto import utils, sbml
from os import listdir
from os.path import isfile, join
from pyasp.asp import *
from xml.etree.ElementTree import ParseError

logger = logging.getLogger(__name__)

###############################################################################
#
message = """
Prepares instance for miscoto. Useful in a benchmark context: pre-calculating
the instance ensures that SBML files do not have to be read again.
Instances are text files that can be modified between runs through multiple
ways, including the use of bash tools
"""

requires = """
requires PyASP package: "pip install PyASP"
"""
#
###############################################################################



def cmd_instance():
    """run miscoto_instance from the shell
    """
    parser = argparse.ArgumentParser(description=message, epilog=requires)
    #parser.add_argument("-h", "--help",
    #                    help="display this message and exit", required=False)
    parser.add_argument("-b", "--bactsymbionts",
                        help="directory of symbionts models, all in sbml format", required=True)
    parser.add_argument("-s", "--seeds",
                        help="seeds in SBML format", required=True)
    parser.add_argument("-m", "--modelhost",
                        help="host metabolic network in SBML format", required=False)
    parser.add_argument("-t", "--targets",
                        help="targets in SBML format", required=False)
    parser.add_argument("-o", "--output",
                        help="output file for instance", required=False)

    args = parser.parse_args()
    bacterium_met =  args.bactsymbionts
    seeds_sbml = args.seeds
    model_host = args.modelhost
    targets_sbml = args.targets
    output = args.output

    run_instance(bacterium_met, seeds_sbml, model_host, targets_sbml, output)

def run_instance(bacteria_dir=None, seeds_file=None, host_file=None, targets_file=None, output=None):
    start_time = time.time()
    if not bacteria_dir or not seeds_file:
        logger.critical("Symbionts and seeds are required minimal inputs")
        sys.exit(1)
    if host_file:
        logger.info('Reading host network from ' + host_file)
        try:
            draftnet = sbml.readSBMLnetwork_symbionts(host_file, 'host_metab_mod')
        except FileNotFoundError:
            logger.critical('Host file not found')
            sys.exit(1)
        except ParseError:
            logger.critical("Invalid syntax in SBML file: "+host_file)
            sys.exit(1)
        draftnet.add(Term('draft', ["\"" + 'host_metab_mod' + "\""]))
    else:
        logger.warning('No host provided')
        draftnet = TermSet()
        draftnet.add(Term('draft', ["\"" + 'host_metab_mod' + "\""]))

    logger.info('Reading seeds from ' + seeds_file)
    try:
        seeds = sbml.readSBMLspecies(seeds_file, 'seed')
    except FileNotFoundError:
        logger.critical('Seeds file not found')
        sys.exit(1)
    except ParseError:
        logger.critical("Invalid syntax in SBML file: "+seeds_file)
        sys.exit(1)
    lp_instance = TermSet(draftnet.union(seeds))

    if targets_file:
        logger.info('Reading targets from ' + targets_file)
        try:
            targets = sbml.readSBMLspecies(targets_file, 'target')
        except FileNotFoundError:
            logger.critical('Targets file not found')
            sys.exit(1)
        except ParseError:
            logger.critical("Invalid syntax in SBML file: "+targets_file)
            sys.exit(1)
        lp_instance = TermSet(lp_instance.union(targets))

    if not os.path.isdir(bacteria_dir):
        logger.critical("Symbiont directory not found")
        sys.exit(1)

    if output:
        #clear the file
        open(output, 'w').close()
        #add the content of seeds and/or targets
        all_networks_file = utils.to_file(lp_instance,output)
    else:
        all_networks_file = utils.to_file(lp_instance)

    logger.info('Reading bacterial networks from ' + bacteria_dir + '...')
    bactfacts = TermSet()
    onlyfiles = [f for f in listdir(bacteria_dir) if isfile(join(bacteria_dir, f))]
    for bacteria_file in onlyfiles:
        name = os.path.splitext(bacteria_file)[0]
        try:
            one_bact_model = sbml.readSBMLnetwork_symbionts(bacteria_dir+'/'+bacteria_file, name)
            one_bact_model.add(Term('bacteria', ["\"" + name + "\""]))
            utils.to_file(one_bact_model, all_networks_file)
            logger.info('Done for ' + name)
        except:
            logger.info('Could not read file ' + name + ' will ignore it')

    logger.info("Instance created: " + os.path.abspath(all_networks_file))

    logger.info("--- %s seconds ---" % (time.time() - start_time))
    utils.clean_up()

    return all_networks_file


if __name__ == '__main__':
    cmd_instance()


# -*- coding: utf-8 -*-
"""
Definition of numerous constants, for paths, names, arguments for ASP solvers.
"""

import sys
import os

# Root
ROOT = __file__.rsplit('/', 1)[0]


# Constants
ASP_FILE_EXTENSION = '.lp'

# Directories (starting from here)
DIR_SOURCES     = ''  # sources are inside the package
DIR_ASP_SOURCES = '/encodings/'
DIR_DATA     = ROOT + '/../data/'  # sources are inside the package

# ASP SOURCES
def __asp_file(name):
    "path to given asp source file name"
    return ROOT + DIR_ASP_SOURCES + name + ASP_FILE_EXTENSION
# Routine
ASP_SRC_SCOPES = __asp_file('scopes')
# Topological subcommunities
ASP_SRC_TOPO_SOUP   = __asp_file('community_soup')
ASP_SRC_TOPO_RXN_MIN_EXCH = __asp_file('community_minexch')
# Topological exchanges
# ASP_SRC_TRANSP   = __asp_file('transported_metabolites')


def basename(filepath):
    """Return the basename of given filepath.
    >>> basename('~/an/interesting/file.lp')
    'file'
    """
    return os.path.splitext(os.path.basename(filepath))[0]

def extension(filepath):
    """Return the extension of given filepath.
    >>> extension('~/an/interesting/file.lp')
    'lp'
    >>> extension('nothing')
    ''
    >>> extension('nothing.important')
    'important'
    """
    return os.path.splitext(os.path.basename(filepath))[1][1:]

def is_valid_path(filepath):
    """True if given filepath is a valid one (a file exists, or could exists)"""
    if filepath and not os.access(filepath, os.W_OK):
        try:
            open(filepath, 'w').close()
            os.unlink(filepath)
            return True
        except OSError:
            return False
    else:  # path is accessible
        return True



#!/usr/bin/python
#-*- coding: utf-8 -*-

import subprocess

from miscoto import run_instance, run_mincom, run_scopes
from pyasp.asp import *


def test_instance():
    expected_lp_instance = set([Term('product',['"e"','"R_R4"','"orgB2"']), Term('product',['"c"','"R_R1"','"host_metab_mod"']), Term('species',['"d"','"d"','"c"','"orgB3"']), Term('species',['"b"','"b"','"c"','"orgB1"']), Term('target',['"f"']), Term('reaction',['"R_R4"','"orgB2"']), Term('product',['"e"','"R_R4"','"orgB3"']), Term('species',['"e"','"e"','"c"','"orgB3"']), Term('species',['"a"','"a"','"c"','"host_metab_mod"']), Term('reactant',['"a"','"R_R1"','"orgB2"']), Term('species',['"e"','"e"','"c"','"orgB2"']), Term('reaction',['"R_R4"','"orgB1"']), Term('reaction',['"R_R1"','"orgB2"']), Term('product',['"c"','"R_R1"','"orgB2"']), Term('reactant',['"d"','"R_R4"','"orgB2"']), Term('species',['"c"','"c"','"c"','"orgB3"']), Term('product',['"f"','"R_R3"','"host_metab_mod"']), Term('reaction',['"R_R3"','"host_metab_mod"']), Term('product',['"c"','"R_R1"','"orgB3"']), Term('reaction',['"R_R2"','"orgB3"']), Term('species',['"d"','"d"','"c"','"orgB1"']), Term('reactant',['"b"','"R_R2"','"orgB3"']), Term('reaction',['"R_R1"','"host_metab_mod"']), Term('reactant',['"b"','"R_R2"','"host_metab_mod"']), Term('species',['"a"','"a"','"c"','"orgB2"']), Term('reactant',['"c"','"R_R4"','"orgB1"']), Term('species',['"e"','"e"','"c"','"host_metab_mod"']), Term('seed',['"b"']), Term('reaction',['"R_R4"','"orgB3"']), Term('species',['"c"','"c"','"c"','"orgB1"']), Term('reactant',['"a"','"R_R1"','"orgB3"']), Term('reactant',['"d"','"R_R4"','"orgB3"']), Term('product',['"d"','"R_R2"','"host_metab_mod"']), Term('species',['"c"','"c"','"c"','"host_metab_mod"']), Term('bacteria',['"orgB3"']), Term('reactant',['"c"','"R_R4"','"orgB2"']), Term('species',['"a"','"a"','"c"','"orgB3"']), Term('species',['"b"','"b"','"c"','"host_metab_mod"']), Term('reactant',['"d"','"R_R4"','"orgB1"']), Term('reactant',['"a"','"R_R1"','"host_metab_mod"']), Term('species',['"d"','"d"','"c"','"host_metab_mod"']), Term('species',['"f"','"f"','"c"','"host_metab_mod"']), Term('reactant',['"e"','"R_R3"','"host_metab_mod"']), Term('product',['"e"','"R_R4"','"orgB1"']), Term('species',['"b"','"b"','"c"','"orgB2"']), Term('product',['"d"','"R_R2"','"orgB3"']), Term('species',['"c"','"c"','"c"','"orgB2"']), Term('species',['"e"','"e"','"c"','"orgB1"']), Term('reaction',['"R_R1"','"orgB3"']), Term('species',['"a"','"a"','"c"','"orgB1"']), Term('reaction',['"R_R2"','"host_metab_mod"']), Term('bacteria',['"orgB1"']), Term('species',['"b"','"b"','"c"','"orgB3"']), Term('draft',['"host_metab_mod"']), Term('species',['"d"','"d"','"c"','"orgB2"']), Term('reactant',['"c"','"R_R4"','"orgB3"']), Term('seed',['"a"']), Term('bacteria',['"orgB2"'])])

    lp_instance = run_instance(host_file='../toy/orgA.xml', bacteria_dir='../toy/symbionts/', seeds_file='../toy/seeds.xml', targets_file='../toy/targets_A.xml')
    lp_instance = [atom for atom in lp_instance]

    assert set(lp_instance) == expected_lp_instance


def test_mincom():
    expected_newly_productible = set(['"f"'])
    expected_bacteria = set(['"orgB3"'])
    expected_exchande = {('"orgB3"', '"host_metab_mod"'): ['"e"']}

    results = run_mincom(host_file='../toy/orgA.xml', bacteria_dir='../toy/symbionts/', seeds_file='../toy/seeds.xml', targets_file='../toy/targets_A.xml', option='minexch')

    assert set(results['newly_prod']) == expected_newly_productible
    assert set(results['bacteria']) == expected_bacteria
    assert results['exchanged'] == expected_exchande


def test_scopes():
    producible_targets = set()
    unproducible_targets = set(['"f"'])
    host_scope = set(['"d"', '"a"', '"c"', '"b"'])
    microbiome_producible_targets = set(['"f"'])
    microbiome_unproducible_targets = set()
    microbiome_only = set(['"f"', '"e"'])

    results = run_scopes(host_file='../toy/orgA.xml', bacteria_dir='../toy/symbionts/', seeds_file='../toy/seeds.xml', targets_file='../toy/targets_A.xml')

    assert set(results['host_prodtargets']) == producible_targets
    assert set(results['host_unprodtargets']) == unproducible_targets
    assert set(results['host_scope']) == host_scope
    assert set(results['com_prodtargets']) == microbiome_producible_targets
    assert set(results['com_unprodtargets']) == microbiome_unproducible_targets
    assert set(results['comhost_scope']) == microbiome_only

if __name__ == "__main__":
    print("** testing scope **")
    test_scopes()
    print("** testing mincom **")
    test_mincom()
    print("** testing instance creation")
    test_instance()



#!/usr/bin/env python
# Copyright (c) 2018, Clemence Frioux <clemence.frioux@inria.fr>
#
# This file is part of miscoto.
#
# miscoto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# miscoto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with miscoto.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-

import argparse
import sys
import os
import time
import logging
from miscoto import query, sbml, commons, utils
from os import listdir
from os.path import isfile, join
from pyasp.asp import *
from xml.etree.ElementTree import ParseError

logger = logging.getLogger(__name__)

###############################################################################
#
message = """
Compute a community from a microbiome
Inputs: SBML models (symbionts and optionally host) + seeds + targets or an
instance pre-created with miscoto_instance.py,
option: soup = minimal size community in a mixed-bag
framework or minexch = minimal size and minimal exchange community.
Can compute one minimal solution and or union, intersection, enumeration of
all minimal solutions
"""

pusage = """
**1** from SBML files
python miscoto_mincom.py -m host.sbml -b symbiont_directory -s seeds.sbml -t targets.sbml -o option [--intersection] [--union] [--enumeration] [--optsol]
\n
**2** from a pre-computed instance with possibly (additional) seeds or targets
python miscoto_mincom.py -a instance.lp -o option [-s seeds.sbml] [-t targets.sbml] [--intersection] [--union] [--enumeration] [--optsol]
\n
Option -o is either 'soup' or 'minexch' depending on the wanted modeling method
\n
"""

requires = """
requires PyASP package: "pip install PyASP"
"""
#
###############################################################################


def cmd_mincom():
    """run directly miscoto_mincom from the shell
    """
    parser = argparse.ArgumentParser(description=message, usage=pusage, epilog=requires)
    parser.add_argument("-o", "--option",
                        help="subcom option: soup, minexch", required=True)
    parser.add_argument("-a", "--asp",
                        help="instance if already created with miscoto_instance", required=False)
    parser.add_argument("--enumeration",
                        help="enumeration of optimal solutions", required=False, action="store_true")
    parser.add_argument("--intersection",
                        help="intersection of optimal solutions", required=False, action="store_true")
    parser.add_argument("--optsol",
                        help="one optimal solutions", required=False, action="store_true")
    parser.add_argument("--union",
                        help="union of optimal solutions", required=False, action="store_true")
    parser.add_argument("-m", "--modelhost",
                        help="host model in SBML format, ignored if -a instance is provided",
                        required=False)
    parser.add_argument("-b", "--bactsymbionts",
                        help="directory of symbionts models, all in sbml format, ignored if -a instance is provided",
                        required=False)
    parser.add_argument("-s", "--seeds",
                        help="seeds in SBML format",
                        required=False)
    parser.add_argument("-t", "--targets",
                        help="targets in SBML format",
                        required=False)

    args = parser.parse_args()
    bacterium_met =  args.bactsymbionts
    option = args.option
    lp_instance_file = args.asp
    targets_sbml = args.targets
    seeds_sbml = args.seeds
    draft_sbml = args.modelhost
    if args.intersection:
        intersection_arg = True
    else:
        intersection_arg = False
    if args.enumeration:
        enumeration_arg = True
    else:
        enumeration_arg = False
    if args.union:
        union_arg = True
    else:
        union_arg = False
    if args.optsol:
        optsol = True
    else:
        optsol = False

    run_mincom(option, bacterium_met, lp_instance_file, targets_sbml, seeds_sbml, draft_sbml,
                intersection_arg, enumeration_arg, union_arg, optsol)

def run_mincom(option=None, bacteria_dir=None, lp_instance_file=None, targets_file=None, seeds_file=None, host_file=None,
                intersection=False, enumeration=False, union=False, optsol=False):
    """Computes community selections in microbiota
        option (str, optional): Defaults to None. Modeling type: 'soup' for uncompartmentalized, 'minexch' for compartmentalized
        bacteria_dir (str, optional): Defaults to None. directory with symbionts metabolic networks
        lp_instance_file (str, optional): Defaults to None. ASP instance file
        targets_file (str, optional): Defaults to None. targets file
        seeds_file (str, optional): Defaults to None. seeds file
        host_file (str, optional): Defaults to None. host metabolic network file
        intersection (bool, optional): Defaults to False. compute intersection of solutions
        enumeration (bool, optional): Defaults to False. compute enumeration of solutions
        union (bool, optional): Defaults to False. compute union of solutions
        optsol (bool, optional): Defaults to False. compute one optimal solution
    """
    start_time = time.time()
    results = {}
    # checking option
    if option == "soup":
        encoding = commons.ASP_SRC_TOPO_SOUP
    elif option == "minexch":
        encoding = commons.ASP_SRC_TOPO_RXN_MIN_EXCH
    else:
        logger.critical("invalid option choice")
        logger.info(pusage)
        quit()

    # case 1: instance is provided, just read targets and seeds if given
    if lp_instance_file:
        if not os.path.isfile(lp_instance_file) :
            logger.critical('Instance file not found')
            sys.exit(1)

        delete_lp_instance = False

        logger.info(
            "Instance provided, only seeds and targets will be added if given")
        if targets_file:
            logger.info('Reading targets from ' + targets_file)
            try:
                targetsfacts = sbml.readSBMLspecies(targets_file, 'target')
            except FileNotFoundError:
                logger.critical('Targets file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: "+targets_file)
                sys.exit(1)
        else:
            targetsfacts = TermSet()

        if seeds_file:
            logger.info('Reading targets from ' + seeds_file)
            try:
                seedsfacts = sbml.readSBMLspecies(seeds_file, 'seed')
            except FileNotFoundError:
                logger.critical('Seeds file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: "+seeds_file)
                sys.exit(1)
        else:
            seedsfacts = TermSet()

        with open(lp_instance_file, "a") as f:
            for elem in targetsfacts:
                f.write(str(elem) + '.\n')
            for elem in seedsfacts:
                f.write(str(elem) + '.\n')

    # case 2: read inputs from SBML files
    elif bacteria_dir and seeds_file and targets_file:
        if not os.path.isdir(bacteria_dir):
            logger.critical("Symbiont directory not found")
            sys.exit(1)

        delete_lp_instance = True

        if host_file:
            logger.info('Reading host network from ' + host_file)
            try:
                draftnet = sbml.readSBMLnetwork_symbionts(host_file, 'host_metab_mod')
            except FileNotFoundError:
                logger.critical('Host file not found')
                sys.exit(1)
            except ParseError:
                logger.critical("Invalid syntax in SBML file: "+host_file)
                sys.exit(1)
            draftnet.add(Term('draft', ["\"" + 'host_metab_mod' + "\""]))
        else:
            logger.warning('No host provided')
            draftnet = TermSet()
            draftnet.add(Term('draft', ["\"" + 'host_metab_mod' + "\""]))

        logger.info('Reading seeds from ' + seeds_file)
        try:
            seeds = sbml.readSBMLspecies(seeds_file, 'seed')
        except FileNotFoundError:
            logger.critical('Targets file not found')
            sys.exit(1)
        except ParseError:
            logger.critical("Invalid syntax in SBML file: "+seeds_file)
            sys.exit(1)
        lp_instance = TermSet(draftnet.union(seeds))

        logger.info('Reading targets from '+ targets_file)
        try:
            targets = sbml.readSBMLspecies(targets_file, 'target')
        except FileNotFoundError:
            logger.critical('Targets file not found')
            sys.exit(1)
        except ParseError:
            logger.critical("Invalid syntax in SBML file: "+targets_file)
            sys.exit(1)
        lp_instance = TermSet(lp_instance.union(targets))

        lp_instance_file = utils.to_file(lp_instance)

        logger.info('Reading bacterial networks from ' + bacteria_dir + '...')
        bactfacts = TermSet()
        onlyfiles = [f for f in listdir(bacteria_dir) if isfile(join(bacteria_dir, f))]
        for bacteria_file in onlyfiles:
            name = os.path.splitext(bacteria_file)[0]
            try:
                one_bact_model = sbml.readSBMLnetwork_symbionts(bacteria_dir+'/'+bacteria_file, name)
                one_bact_model.add(Term('bacteria', ["\"" + name + "\""]))
                utils.to_file(one_bact_model, lp_instance_file)
                logger.info('Done for ' + name)
            except:
                logger.info('Could not read file ' + name + ' will ignore it')

    else:
        logger.info(
            "ERROR missing input: missing instance or symbionts/targets/seeds")
        logger.info(pusage)
        quit()

    if not optsol and not union and not enumeration and not intersection:
        logger.info(
            "No choice of solution provided. Will compute one optimal solution by default"
        )
        optsol = True

    logger.info('\nFinding optimal communities for target production...')
    #ground the instance
    try:
        grounded_instance = query.get_grounded_communities_from_file(lp_instance_file, encoding)
    except OSError:
        logger.critical(
            "Error. Possible reason: solvers are not properly installed. Please install them again by running 'pip uninstall pyasp' and 'pip install pyasp --no-cache-dir'"
        )
        sys.exit(1)


# one solution
    if optsol:
        logger.info('\n*** ONE MINIMAL SOLUTION ***')
        one_model = query.get_communities_from_g(grounded_instance)
        optimum = ','.join(map(str, one_model.score))
        still_unprod = []
        bacteria = []
        newly_prod = []
        exchanged = {}
        for a in one_model:
            if a.pred() == 'unproducible_target':
                still_unprod.append(a.arg(0))
            elif a.pred() == 'newly_producible_target':
                newly_prod.append(a.arg(0))
            elif a.pred() == 'chosen_bacteria':
                bacteria.append(a.arg(0))
            elif a.pred() == 'exchanged':
                if (a.arg(2),a.arg(3)) in exchanged: #exchanged[(from,to)]=[(what,compartto);(what,compartto)]
                    exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
                else:
                    exchanged[(a.arg(2),a.arg(3))] = []
                    exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
        logger.info(str(len(newly_prod)) + ' newly producible target(s):')
        logger.info("\n".join(newly_prod))
        logger.info('Still ' + str(len(still_unprod)) + ' unproducible target(s):')
        logger.info("\n".join(still_unprod))
        logger.info('Minimal set of bacteria of size ' + str(len(bacteria)))
        logger.info("\n".join(bacteria))
        if len(exchanged) >= 1:
            logger.info('Minimal set of exchanges of size => ' +
                        str(sum(len(v) for v in exchanged.values())))
            for fromto in exchanged:
                logger.info("\texchange(s) from " + fromto[0] + ' to ' +
                            fromto[1] + " = " + ','.join(exchanged[fromto]))
        results['one_model'] = one_model
        results['exchanged'] = exchanged
        results['bacteria'] = bacteria
        results['still_unprod'] = still_unprod
        results['newly_prod'] = newly_prod

# union of solutions
    if union:
        logger.info('\n*** UNION OF MINIMAL SOLUTION ***')
        try:
            if optsol:
                union_m = query.get_union_communities_from_g(grounded_instance, optimum)
            else:
                union_m = query.get_union_communities_from_g_noopti(grounded_instance)
        except IndexError:
            logger.error(
                "No stable model was found. Possible troubleshooting: no harmony between names for identical metabolites among host and microbes"
            )
            quit()
        optimum_union = ','.join(map(str, union_m.score))
        union_bacteria = []
        union_exchanged = {}
        for a in union_m :
            if a.pred() == 'chosen_bacteria':
                union_bacteria.append(a.arg(0))
            elif a.pred() == 'exchanged':
                if (a.arg(2),a.arg(3)) in union_exchanged: #union_exchanged[(from,to)]=[(what,compartto);(what,compartto)]
                    union_exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
                else:
                    union_exchanged[(a.arg(2),a.arg(3))] = []
                    union_exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
        logger.info('Union of minimal sets of bacteria, with optimum = ' +
                    optimum_union + ' comprises ' + str(len(union_bacteria)) +
                    ' bacteria')
        logger.info("\n".join(union_bacteria))
        if len(union_exchanged) >= 1:
            logger.info('\nExchanges in union => ' +
                        str(sum(len(v) for v in union_exchanged.values())))
            for fromto in union_exchanged:
                logger.info('\texchange(s) from ' + fromto[0] + ' to ' +
                            fromto[1] + " = " +
                            ','.join(union_exchanged[fromto]))
        results['union_exchanged'] = union_exchanged
        results['union_bacteria'] = union_bacteria
        results['optimum_union'] = optimum_union

# intersection of solutions
    if intersection:
        logger.info('\n*** INTERSECTION OF MINIMAL SOLUTION ***')
        if optsol:
            intersection_m = query.get_intersection_communities_from_g(grounded_instance, optimum)
        else:
            intersection_m = query.get_intersection_communities_from_g_noopti(grounded_instance)
        optimum_inter = ','.join(map(str, intersection_m.score))
        inter_bacteria = []
        inter_exchanged = {}
        for a in intersection_m :
            if a.pred() == 'chosen_bacteria':
                inter_bacteria.append(a.arg(0))
            elif a.pred() == 'exchanged':
                if (a.arg(2),a.arg(3)) in inter_exchanged: #inter_exchanged[(from,to)]=[(what,compartto);(what,compartto)]
                    inter_exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
                else:
                    inter_exchanged[(a.arg(2),a.arg(3))] = []
                    inter_exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
        logger.info('Intersection of minimal sets of bacteria, with optimum = '
                    + optimum_inter + ' comprises ' +
                    str(len(inter_bacteria)) + ' bacteria')
        logger.info("\n".join(inter_bacteria))
        if len(inter_exchanged) >= 1:
            logger.info('\nExchanges in intersection => ' +
                        str(sum(len(v) for v in inter_exchanged.values())))
            for fromto in inter_exchanged:
                logger.info('\texchange(s) from ' + fromto[0] + ' to ' +
                            fromto[1] + " = " +
                            ','.join(inter_exchanged[fromto]))
        results['inter_exchanged'] = inter_exchanged
        results['inter_bacteria'] = inter_bacteria
        results['optimum_inter'] = optimum_inter

# enumeration of all solutions
    if enumeration:
        logger.info('\n*** ENUMERATION OF MINIMAL SOLUTION ***')
        if optsol:
            all_models = query.get_all_communities_from_g(grounded_instance, optimum)
        else:
            all_models = query.get_all_communities_from_g_noopti(grounded_instance)
        count = 1
        for model in all_models:
            enum_bacteria = []
            enum_exchanged = {}
            logger.info('\nSolution ' + str(count))
            for a in model :
                if a.pred() == 'chosen_bacteria':
                    enum_bacteria.append(a.arg(0))
                elif a.pred() == 'exchanged':
                    if (a.arg(2),a.arg(3)) in enum_exchanged: #enum_exchanged[(from,to)]=[(what,compartto);(what,compartto)]
                        enum_exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
                    else:
                        enum_exchanged[(a.arg(2),a.arg(3))] = []
                        enum_exchanged[(a.arg(2),a.arg(3))].append(a.arg(0))
            logger.info("\t" + str(len(enum_bacteria)) +
                        " bacterium(ia) in solution " + str(count))
            for elem in enum_bacteria:
                logger.info("\t" + elem)
            if len(enum_exchanged) >= 1:
                logger.info("\t" +
                            str(sum(len(v) for v in enum_exchanged.values())) +
                            " exchange(s) in solution " + str(count))
                for fromto in enum_exchanged:
                    logger.info('\texchange(s) from ' + fromto[0] + ' to ' +
                                fromto[1] + " = " +
                                ','.join(enum_exchanged[fromto]))
            count+=1
        results['all_models'] = all_models
        results['enum_exchanged'] = enum_exchanged
        results['enum_bacteria'] = enum_bacteria

    if delete_lp_instance == True:
        os.unlink(lp_instance_file)

    logger.info("--- %s seconds ---" % (time.time() - start_time))
    utils.clean_up()

    return results

if __name__ == '__main__':
    cmd_mincom()


#!python
# -*- coding: utf-8 -*-
import os
from pyasp.term import *

def clean_up() :
    if os.path.isfile("parser.out"): os.remove("parser.out")
    if os.path.isfile("parsetab.py"): os.remove("parsetab.py")
    if os.path.isfile("asp_py_lextab.py"): os.remove("asp_py_lextab.py")
    if os.path.isfile("asp_py_lextab.pyc"): os.remove("asp_py_lextab.pyc")
    if os.path.isfile("asp_py_parsetab.py"): os.remove("asp_py_parsetab.py")
    if os.path.isfile("asp_py_parsetab.pyc"): os.remove("asp_py_parsetab.pyc")

def to_file(termset, outputfile=None):
    """write the content of the TermSet into a file
    
    Args:
        termset (TermSet): ASP termset
        outputfile (str, optional): Defaults to None. name of the output file
    """
    if outputfile:
        f = open(outputfile, 'a')
    else:
        fd, outputfile = tempfile.mkstemp(suffix='.lp', prefix='miscoto_')
        f = os.fdopen(fd, 'a')
    for t in termset:
        f.write(str(t) + '.\n')
    f.close()
    return outputfile
